//---------------------------------------------------------------------------
#include "stm32f4xx.h"
#include <stdio.h>
//---------------------------------------------------------------------------
#define display_on      0x0C      //Cursor OFF, Blink off
#define display_off     0x08
#define cursor_on       0x0E
#define cursor_blink    0x0F
#define home            0x02
#define clear           0x01
#define auto_increment  0x06
#define auto_decrement  0x04
#define function_set    0x28       //DL = 0 <Data Lenghth> (4 bit mode). N = 1 <Num lines> (2 Lines).
#define font2           function_set | 0x04     //Set bit 3 only.
//---------------------------------------------------------------------------
void LCD_pinDirs(void);
void LCD_init(void);
void LCD_transmit(char data);
void LCD_data(char data);
void LCD_command(char command);
void LCD_string(char row, char column, char text[]);
void LCD_getBits(char pinBit[], char data);
void LCD_sendBits(char pinBit[]);
void latch(void);
//---------------------------------------------------------------------------
char  rowAddr[] = {0x80,0xC0,0x94,0xD4};  
//---------------------------------------------------------------------------
void HSI_init(void);
void delay_ms(int n); 
//---------------------------------------------------------------------------
int main()
{	  
	HSI_init();
	LCD_init();
	
	while(1)
	{
		LCD_string(0,0, "Bugz Bensce!");
		LCD_string(1,3, "ARM 2024");
	}
}
//---------------------------------------------------------------------------
void HSI_init(void)  //Set Clock to 16Mhz
{
	// Set the HSION bit and wait for HSIRDY
	RCC->CR |= 1ul << 0;
	while(!(RCC->CR & 2ul)){};

  // System clock switch to HSI		
  RCC->CFGR &= ~(3ul << 0);   // HSI oscillator selected as system clock
	RCC->CFGR &= ~(7ul << 27);  // MCO2 Prescaler. NO DIVISION selected 
	RCC->CFGR &= ~(3ul << 30);  // System clock (SYSCLK) selected
}
//--------------------------------------------------------------------------
void LCD_pinDirs(void)
{
	//SYSCFGEN: System configuration controller clock enable 
	RCC->APB2ENR |= RCC_APB2ENR_SYSCFGEN;
	
	//GPIOA Clock Enable
	RCC->AHB1ENR |= RCC_AHB1ENR_GPIOAEN;

	//Configure PA0 as output D4
	GPIOA->MODER &= ~0x00000003;
	GPIOA->MODER |= 0x00000001;

	//Configure PA1 as output D5
	GPIOA->MODER &= ~0x00000000C;
	GPIOA->MODER |= 0x00000004;

	//Configure PA2 as output D6
	GPIOA->MODER &= ~0x000000030;
	GPIOA->MODER |= 0x00000010;

	//Configure PA3 as output D7
	GPIOA->MODER &= ~0x0000000C0;
	GPIOA->MODER |= 0x00000040;

	//Configure PA5 as output EN
	GPIOA->MODER &= ~0x00000C00;
	GPIOA->MODER |= 0x00000400;

	//Configure PA6 as output RS
	GPIOA->MODER &= ~0x00003000;
	GPIOA->MODER |= 0x00001000;
}
//-----------------------Follow Datasheet instructions--------------------------
void LCD_init()
{
  LCD_pinDirs();

  LCD_command(home);
	LCD_command(function_set);   
	LCD_command(display_on);	
	LCD_command(auto_increment);
}
//------------------------------------------------------------------------------
void LCD_command(char command)
{
  GPIOA->BSRR |= GPIO_BSRR_BR6;	 //RS = 0: Instruction register ( commands ). 
  LCD_transmit(command);
}
//------------------------------------------------------------------------------
void LCD_data(char data)
{
	GPIOA->BSRR |= GPIO_BSRR_BS6;  //RS = 1: Data register ( strings and characters ).
  LCD_transmit(data);
}
//------------------------------------------------------------------------------
void LCD_string(char row, char column, char text[])
{
  LCD_command(rowAddr[row] + column);  //Set coordinates.
	
  while(*text)
  {
    LCD_data(*text++);
  }
}
//------------------------------------------------------------------------------
void LCD_transmit(char data)
{
  char pinBits[8];
   
  LCD_getBits(pinBits, data);
  LCD_sendBits(pinBits);
}
//------------------------------------------------------------------------------
void LCD_sendBits(char pinBit[])
{		
	GPIOA->ODR &= 0xFFF0;		//Clear Data Pins (PA0: D4, PA1: D5, PA2: D6, PA3: D7)
	  
	//Higher Nibble
  GPIOA->ODR |= (pinBit[4] << 0);  //D4
  GPIOA->ODR |= (pinBit[5] << 1);  //D5
  GPIOA->ODR |= (pinBit[6] << 2);  //D6
  GPIOA->ODR |= (pinBit[7] << 3);  //D7
  latch();
  
	GPIOA->ODR &= 0xFFF0;		//Clear Data Pins (PA0: D4, PA1: D5, PA2: D6, PA3: D7)
	//Lower Nibble
  GPIOA->ODR |= (pinBit[0] << 0);  //D4
  GPIOA->ODR |= (pinBit[1] << 1);  //D5
  GPIOA->ODR |= (pinBit[2] << 2);  //D6
  GPIOA->ODR |= (pinBit[3] << 3);  //D7
  latch();
}
//------------------------------------------------------------------------------
void LCD_getBits(char pinBit[], char data)
{ 
  for(char i = 0; i < 8; i++)
  {
    pinBit[i] = ((data >> i) & 0x01) ? 1 : 0;
  }
}
//------------------------------------------------------------------------------
void latch()
{
	GPIOA->BSRR |= GPIO_BSRR_BS5; //EN = 1
	delay_ms(5);
	GPIOA->BSRR |= GPIO_BSRR_BR5; //EN = 0
}
//--------------------------------------------------------------------------
void delay_ms(int n) 
{
  int i;
  for(; n > 0; n--)
     for (i = 0; i < 3195; i++);
}
//--------------------------------------------------------------------------
